from systemrdl.node import AddressableNode, AddrmapNode, Node
from .state import State

def get_node_prefix(ds: State, root_node: AddrmapNode, node: AddressableNode) -> str:
    # if ds.reuse_typedefs:
    #     prefix = node.get_global_type_name("__")
    #     if prefix is None:
    #         # Unable to determine a reusable type name. Fall back to hierarchical path
    #         # Add prefix to prevent collision when mixing namespace methods
    #         prefix = "xtern__" + node.get_rel_path(
    #             root_node.parent,
    #             hier_separator="__",
    #             array_suffix="x",
    #             empty_array_suffix="x"
    #         )
    #else:
    prefix = node.get_rel_path(
            root_node.parent,
            hier_separator="__",
            array_suffix="x",
            empty_array_suffix="x"
        )
    return prefix


def get_struct_name(ds: State, root_node: AddrmapNode, node: AddressableNode) -> str:
    # if node.is_array and node.array_stride > node.size:
    #     # Stride is larger than size of actual element.
    #     # Struct will be padded up, and therefore needs a unique name
    #     pad_suffix = f"__stride{node.array_stride:x}"
    # else:
    #     pad_suffix = ""

    return get_node_prefix(ds, root_node, node)


def get_friendly_name(ds: State, root_node: AddrmapNode, node: Node) -> str:
    """
    Returns a useful string that helps identify the typedef in
    a comment
    """
    # if ds.reuse_typedefs:
    #     friendly_name = node.get_global_type_name("::")

    #     if friendly_name is None:
    #         # Unable to determine a reusable type name. Fall back to hierarchical path
    #         friendly_name = node.get_rel_path(root_node.parent)
    # else:
    friendly_name = node.get_rel_path(root_node.parent)

    return type(node.inst).__name__ + " - " + friendly_name

def roundup_pow2(x: int) -> int:
    return 1<<(x-1).bit_length()



from typing import TextIO, Set, Optional, List
import os
import re

from systemrdl.walker import RDLListener, RDLWalker, WalkerAction
from systemrdl.node import AddrmapNode, AddressableNode, RegNode, FieldNode, Node

from .state import State
from .identifier_filter import kw_filter as kwf
from . import utils

class ErlangHeaderGenerator(RDLListener):
    def __init__(self, s: State) -> None:
        self.s = s

        self.defined_namespace: Set[str]
        self.defined_namespace = set()
        self.indent_level = 0

        self.root_node: AddrmapNode
        self.root_node = None

        self.f: TextIO
        self.f = None # type: ignore

    def run(self, base_path: str, top_nodes: List[AddrmapNode]) -> None:
        path = base_path + f".{self.s.flavor.header_extension}"

        context = {
            "module": re.sub(r"[^\w]", "_", os.path.basename(path).split('.')[0]),
        }

        # Stream header via jinja
        output = self.s.jj_env.get_template("header.hrl").render(context)

        with open(path, "w", encoding='utf-8') as f:
            self.f = f

            f.write(output)

            f.write("\n")

            # Generate definitions
            for node in top_nodes:
                self.root_node = node
                RDLWalker().walk(node, self)

            # Ensure newline before EOF
            f.write("\n")

    def push_indent(self) -> None:
        self.indent_level += 1

    def pop_indent(self) -> None:
        self.indent_level -= 1

    def write(self, s: str) -> None:
        if self.indent_level:
            self.f.write("    " * self.indent_level)
        self.f.write(s)

    def get_node_prefix(self, node: AddressableNode) -> str:
        return utils.get_node_prefix(self.s, self.root_node, node)

    def get_struct_name(self, node: AddressableNode) -> str:
        return utils.get_struct_name(self.s, self.root_node, node)

    def get_friendly_name(self, node: Node) -> str:
        return utils.get_friendly_name(self.s, self.root_node, node)

    def write_bitfields(self, grp_name: str, regwidth: int, fields: List[FieldNode]) -> None:
        if not fields:
            return

        pfields = []
        # if regwidth > 64:
        #     # TODO: add support for this
        #     self.root_node.env.msg.fatal(
        #         "C header bit-fields for registers wider than 64-bits is not supported yet",
        #         fields[0].parent.inst.inst_src_ref
        #     )

        if self.s.bitfield_order_ltoh:
            # TODO: double check it
            # Bits are packed in struct LSb --> MSb
            current_offset = 0
            for field in fields:
                # if field.low > current_offset:
                #     self.write(f"uint{regwidth}_t :{field.low - current_offset:d};\n")
                #     current_offset = field.low
                pfields.append(f"    {kwf(field.inst_name.upper())}:{field.width:d}")
                #self.write(f"{kwf(field.inst_name)}:{field.width:d},\n")
                current_offset += field.width

            # if current_offset < regwidth:
            #     self.write(f"uint{regwidth}_t :{regwidth - current_offset:d};\n")
        else:
            # TODO: double check it
            # Bits are packed in struct MSb --> LSb
            current_offset = 0
            for field in fields:
                # if field.low > current_offset:
                #     self.write(f"uint{regwidth}_t :{field.low - current_offset:d};\n")
                #     current_offset = field.low
                pfields.append(f"    {kwf(field.inst_name.upper())}:{field.width:d}")
                #self.write(f"{kwf(field.inst_name)}:{field.width:d},\n")
                current_offset += field.width

        self.write(",\n".join(pfields))

    def write_define_bit_syntax(self, grp_name: str, union_name: str, regwidth: int, fields: List[FieldNode]) -> None:
        if not fields:
            return

        self.write(f"-define({union_name},\n")
        self.write_bitfields(grp_name, regwidth, fields)
        self.write("\n).\n")

    def write_record(self, node: RegNode) -> None:
        union_name = self.get_struct_name(node)
        pfields = []

        self.write(f"-record({union_name}, {{\n")

        for field in node.fields():
            pfields.append(f"    {kwf(field.inst_name.lower())}")

        self.write(",\n".join(pfields))

        self.write("\n}).\n")

    def enter_Reg(self, node: RegNode) -> Optional[WalkerAction]:
        prefix = self.get_node_prefix(node).upper()

        if prefix in self.defined_namespace:
            return WalkerAction.SkipDescendants
        self.defined_namespace.add(prefix)

        self.write(f"\n% {self.get_friendly_name(node)}\n")

        for field in node.fields():
            field_prefix = prefix + "__" + field.inst_name.upper()

            bm = ((1 << field.width) - 1) << field.low
            self.write(f"-define({field_prefix}_bm, 16#{bm:x}).\n")
            self.write(f"-define({field_prefix}_bp, {field.low:d}).\n")
            self.write(f"-define({field_prefix}_bw, {field.width:d}).\n")

            reset = field.get_property('reset')
            if isinstance(reset, int):
                self.write(f"-define({field_prefix}_reset, 16#{reset:x}).\n")

        # No need to traverse fields
        return WalkerAction.SkipDescendants


    def exit_Reg(self, node: RegNode) -> None:
        if not self.s.generate_bitfields:
            return

        union_name = self.get_struct_name(node)
        if union_name in self.defined_namespace:
            # Already defined. Skip
            return
        self.defined_namespace.add(union_name)

        # Sort fields into their respective categories
        overlapping_fields = self.s.overlapping_fields.get(node.get_path(), [])
        fr_fields = []
        fw_fields = []
        f_fields = []
        for field in node.fields():
            if field.inst_name in overlapping_fields:
                # Is an overlapping field.
                # Guaranteed to be either read-only or write-only
                if field.is_sw_readable:
                    fr_fields.append(field)
                else:
                    fw_fields.append(field)
            else:
                f_fields.append(field)

        regwidth = node.get_property('regwidth')

        self.write_define_bit_syntax("f", union_name.upper() + "_f", regwidth, f_fields)
        self.write_define_bit_syntax("fr", union_name.upper() + "_fr", regwidth, fr_fields)
        self.write_define_bit_syntax("fw", union_name.upper() + "_fw", regwidth, fw_fields)

        print(f"{union_name}")

        self.write_record(node)



from typing import Any, Dict, List
import os

import jinja2 as jj
from systemrdl.node import AddrmapNode

from .beam_languages import BeamLanguages

class State:
    def __init__(self, top_node: AddrmapNode, kwargs: Any) -> None:
        loader = jj.FileSystemLoader(os.path.join(os.path.dirname(__file__), "templates"))
        self.jj_env = jj.Environment(
            loader=loader,
            undefined=jj.StrictUndefined
        )

        self.top_node = top_node

        #------------------------
        # Info about the design
        #------------------------
        # Each reg that has overlapping fields generates an entry:
        #   reg_path : list of field names involved in overlap
        self.overlapping_fields = {} # type: Dict[str, List[str]]

        # Pairs of overlapping registers
        #   first_reg_path : partner_register_name
        self.overlapping_reg_pairs = {} # type: Dict[str, str]

        #------------------------
        # Extract compiler args
        #------------------------
        self.flavor: BeamLanguages
        self.flavor = kwargs.pop("flavor", BeamLanguages.default)
        assert isinstance(self.flavor, BeamLanguages)

        # self.reuse_typedefs: bool
        # self.reuse_typedefs = kwargs.pop("reuse_typedefs", True)

        # # Enable generation of bit-field structs for registers
        self.generate_bitfields: bool
        self.generate_bitfields = kwargs.pop("generate_bitfields", True)

        # # Bitfield order is implementation defined
        self.bitfield_order_ltoh: bool
        self.bitfield_order_ltoh = kwargs.pop("bitfield_order_ltoh", True)

        # # If a register is wider than 64-bits, it cannot be represented by a stdint
        # # type. Therefore it must be represented by an array of subwords
        # self.wide_reg_subword_size: int
        # self.wide_reg_subword_size = kwargs.pop("wide_reg_subword_size", 32)
        # assert self.wide_reg_subword_size in {8, 16, 32, 64}

        self.explode_top: bool
        self.explode_top = kwargs.pop("explode_top", False)

        # self.instantiate: bool
        # self.instantiate = kwargs.pop("instantiate", False)

        # self.inst_offset: int
        # self.inst_offset = kwargs.pop("inst_offset", 0)

        # self.testcase: bool
        # self.testcase = kwargs.pop("testcase", False)




ERL_KEYWORDS = {
    "case", "receive", "send", "maybe", "if"
}

def kw_filter(s: str) -> str:
    """
    Make all user identifiers 'safe' and ensure they do not collide with
    Erlang keywords.

    If an Erlang keyword is encountered, make it an atom
    """
    if s in ERL_KEYWORDS:
        s = f"'{s}'"
    return s



from typing import Optional, List

from systemrdl.walker import RDLListener, RDLWalker, WalkerAction
from systemrdl.node import AddrmapNode, RegNode, AddressableNode, FieldNode

from .state import State

class Scanner(RDLListener):
    def __init__(self, s: State) -> None:
        self.s = s
        self.msg = s.top_node.env.msg

        self.prev_reg_stack: List[RegNode]
        self.prev_reg_stack = []

        self.indent = 0

    @property
    def top_node(self) -> AddrmapNode:
        return self.s.top_node

    def run(self) -> None:
        RDLWalker().walk(self.top_node, self)
        if self.msg.had_error:
            self.msg.fatal(
                "Unable to export due to previous errors"
            )

    def enter_AddressableComponent(self, node: AddressableNode) -> Optional[WalkerAction]:
        if not isinstance(node, RegNode):
            self.prev_reg_stack.append(None)
        return WalkerAction.Continue

    def exit_AddressableComponent(self, node: AddressableNode) -> Optional[WalkerAction]:
        if not isinstance(node, RegNode):
            self.prev_reg_stack.pop()
        return WalkerAction.Continue

    def enter_Component(self, node):
        if not isinstance(node, FieldNode):
            print("\t"*self.indent, node.get_path_segment())
            self.indent += 1

    def enter_Field(self, node):
        # Print some stuff about the field
        bit_range_str = "[%d:%d]" % (node.high, node.low)
        sw_access_str = "sw=%s" % node.get_property('sw').name
        print("\t"*self.indent, bit_range_str, node.get_path_segment(), sw_access_str)

    def exit_Component(self, node):
        if not isinstance(node, FieldNode):
            self.indent -= 1

    def enter_Reg(self, node: RegNode) -> Optional[WalkerAction]:

        # Collect information about overlapping fields, if any.
        overlapping_fields = []
        fields = list(node.fields())
        reg_bitmask = 0
        for i, field in enumerate(fields):
            field_bitmask = ((1 << field.width) - 1) << field.low
            if field_bitmask & reg_bitmask:
                # this field overlaps with a prior one
                # Determine which one
                for prior_field in fields[0:i]:
                    if prior_field.high >= field.low:
                        if prior_field.inst_name not in overlapping_fields:
                            overlapping_fields.append(prior_field.inst_name)

                if field.inst_name not in overlapping_fields:
                    overlapping_fields.append(field.inst_name)

            reg_bitmask |= field_bitmask
        if overlapping_fields:
            # Save infor about this register for later.
            self.s.overlapping_fields[node.get_path()] = overlapping_fields


        # Check previous adjacent register for overlap
        prev_reg = self.prev_reg_stack[-1]
        if prev_reg and ((prev_reg.raw_address_offset + prev_reg.total_size) > node.raw_address_offset):
            # registers overlap!

            # Registers shall be co-located.
            # This restriction guarantees that overlaps can only happen in pairs,
            # and avoids the more complex overlap scenarios that involve multiple registers.
            if (
                prev_reg.raw_address_offset != node.raw_address_offset # Same offset
                or prev_reg.size != node.size # Same size
                or prev_reg.total_size != node.total_size # Same array footprint
            ):
                self.msg.error(
                    "C header export currently only supports registers that are co-located. "
                    f"See registers: '{prev_reg.inst_name}' and '{node.inst_name}.'",
                    node.inst.inst_src_ref
                )

            # Save information about register overlap pair
            self.s.overlapping_reg_pairs[prev_reg.get_path()] = node.inst_name

        # Check for sparse register arrays
        if node.is_array and node.array_stride > node.size:
            self.msg.error(
                "C header export does not support sparse arrays of registers. "
                f"See register: '{node.inst_name}.'",
                node.inst.inst_src_ref
            )

        return WalkerAction.SkipDescendants

    def exit_Reg(self, node: RegNode) -> None:
        self.prev_reg_stack[-1] = node



from typing import TYPE_CHECKING

from peakrdl.plugins.exporter import ExporterSubcommandPlugin #pylint: disable=import-error
from peakrdl.config import schema #pylint: disable=import-error

from .exporter import ErlangExporter
from .beam_languages import BeamLanguages

if TYPE_CHECKING:
    import argparse
    from systemrdl.node import AddrmapNode

class Exporter(ExporterSubcommandPlugin):
    short_desc = "Generate Erlang and Elixir code of an address space"

    cfg_schema = {
         "flavor": schema.Choice(list(BeamLanguages.__members__.keys())),
    #     "type_style": schema.Choice(['lexical', 'hier']),
    #     "subword_size": schema.Integer(),
        "bitfields": schema.Choice(["ltoh", "htol", "none"]),
    }

    def add_exporter_arguments(self, arg_group: 'argparse._ActionsContainer') -> None:
        arg_group.add_argument(
            "--flavor",
            choices=list(BeamLanguages.__members__.keys()),
            default=None,
            help=f"""
            Select the C standard that generated output will conform to. [{BeamLanguages.default.name}]
            """
        )

        arg_group.add_argument(
            "-b", "--bitfields",
            choices=["ltoh", "htol", "none"],
            default=None,
            help="""
            Enable generation of register bitfield structs to provide bit-level access
            to fields.

            Since the packing order of C struct bitfields is implementation defined, the
            packing order must be explicitly specified. [none]
            """
        )

        # arg_group.add_argument(
        #     "-x", "--explode-top",
        #     action="store_true",
        #     default=False,
        #     help=""""
        #     If set, the top-level hiearchy is skipped. Instead, definitions for
        #     all the direct children are generated.

        #     Note that only block-like definitons are generated.
        #     i.e: children that are registers are skipped.
        #     """
        # )

        # arg_group.add_argument(
        #     "-i", "--instantiate",
        #     action="store_true",
        #     default=False,
        #     help=""""
        #     If set, header will also include a macro that instantiates each top-level
        #     block at a defined hardware address, allowing for direct access.
        #     """
        # )

        # Wrap constructor to allow hex strings
        # def integer(n):
        #     return int(n, 0)

        # arg_group.add_argument(
        #     "--inst-offset",
        #     type=integer,
        #     default=0,
        #     help="""
        #     Apply an additional address offset to instance definitions.
        #     """
        # )

        # arg_group.add_argument(
        #     "--type-style",
        #     dest="type_style",
        #     choices=['lexical', 'hier'],
        #     default=None,
        #     help="""Choose how typedef names are generated.
        #     The 'lexical' style will use RDL lexical scope & type names where
        #     possible and attempt to re-use equivalent type definitions.
        #     The 'hier' style uses component's hierarchy as the struct type name. [lexical]
        #     """
        # )

        # arg_group.add_argument(
        #     "--subword-size",
        #     type=int,
        #     default=None,
        #     help="""
        #     C's <stdint.h> types only extend up to 64-bit types.

        #     If a register is encountered that is larger than this, the generated
        #     header will represent it using an array of smaller sub-words.
        #     Set the desired sub-word size of 8, 16, 32 or 64. [32]
        #     """
        # )

        # arg_group.add_argument(
        #     "--testcase",
        #     action="store_true",
        #     default=False,
        #     help="""
        #     Create a testcase C file that validates the header
        #     """
        # )

    def do_export(self, top_node: 'AddrmapNode', options: 'argparse.Namespace') -> None:
        flavor_name = options.flavor or self.cfg['flavor'] or "default"
        flavor = BeamLanguages[flavor_name]

        bitfields = options.bitfields or self.cfg['bitfields'] or "none"
        generate_bitfields = bitfields != "none"
        bitfield_order_ltoh = bitfields == "ltoh"

        # type_style = options.type_style or self.cfg['type_style'] or "lexical"
        # reuse_typedefs = type_style == "lexical"

        # subword_size = options.subword_size or self.cfg['subword_size'] or 32

        # x = CHeaderExporter()
        # x.export(
        #     top_node,
        #     path=options.output,
        #     std=std,
        #     generate_bitfields=generate_bitfields,
        #     bitfield_order_ltoh=bitfield_order_ltoh,
        #     reuse_typedefs=reuse_typedefs,
        #     wide_reg_subword_size=subword_size,
        #     explode_top=options.explode_top,
        #     instantiate=options.instantiate,
        #     inst_offset=options.inst_offset,
        #     testcase=options.testcase,
        # )
        x = ErlangExporter()
        x.export(top_node, path=options.output, flavor=flavor, bitfield_order_ltoh=bitfield_order_ltoh)



from typing import Dict
import enum

class BeamLanguages(enum.Enum):

    def __init__(self, name: str, d: Dict[str, str]) -> None:
        self.module_extension = d.get("module_extension")
        self.header_extension = d.get("header_extension")

        # Prevent Enum from flattening members into aliases
        self._value_ = name

    erlang = "erlang", {
        "module_extension": "erl",
        "header_extension": "hrl"
    }

    elixir = "elixir", {
        "module_extension": "ex",
        "header_extension": ""
    }

    default = erlang



import os

from typing import Any, Union

from systemrdl.node import RootNode, AddrmapNode, MemNode, RegfileNode

from .scanner import Scanner
from .state import State
from .erlang_module import ErlangModuleGenerator
from .erlang_header import ErlangHeaderGenerator

class ErlangExporter:
    def export(self, node: Union[RootNode, AddrmapNode], path: str, **kwargs: Any) -> None:
        """
        Parameters
        ----------
        node: AddrmapNode
            Top-level SystemRDL node to export.
        path: str
            Output header file path
        flavor: BeamLanguage
            Select which supported BEAM language to use
        bitfield_order_ltoh: bool
            The packing order of C struct bitfields is implementation defined.
            If True, packing will assume low-to-high bit-packing order.
        """
        # If it is the root node, skip to top addrmap
        if isinstance(node, RootNode):
            top_node = node.top
        else:
            top_node = node

        s = State(top_node, kwargs)

        # Check for stray kwargs
        if kwargs:
            raise TypeError(f"got an unexpected keyword argument '{list(kwargs.keys())[0]}'")

        # Validate and collect info for export
        Scanner(s).run()

        top_nodes = []
        if s.explode_top:
            for child in top_node.children():
                if isinstance(child, (AddrmapNode, MemNode, RegfileNode)):
                    top_nodes.append(child)
        else:
            top_nodes.append(top_node)

        (base_path, ext) = os.path.splitext(path)

        # Write output
        ErlangModuleGenerator(s).run(base_path, top_nodes)
        ErlangHeaderGenerator(s).run(base_path, top_nodes)
